//! Machine-level Intermediate Representation

use std::collections::btree_map::Entry;
use std::collections::{BTreeMap, BTreeSet, HashMap, HashSet};

use itertools::Itertools;

use crate::string_table::{Index as StringsIndex, StringTable};

#[derive(Debug, PartialEq, Eq, Clone, Copy, Hash)]
pub enum Type {
    /// 8bit integer
    Byte,
    /// 16bit integer
    Word,
    /// 32bit integer
    DWord,
    /// 64bit integer
    QWord,
    /// 32bit ieee-754
    SinglePrecision,
    /// 64bit ieee-754
    DoublePrecision,
    // XWord,
    // YWord,
    // ZWord,
}

impl core::fmt::Display for Type {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let name = match self {
            Type::Byte => "byte",
            Type::Word => "word",
            Type::DWord => "dword",
            Type::QWord => "qword",
            Type::SinglePrecision => "f32",
            Type::DoublePrecision => "f64",
        };

        write!(f, "{name}")
    }
}

impl Type {
    pub const fn is_floating(self) -> bool {
        match self {
            Type::SinglePrecision | Type::DoublePrecision => true,
            _ => false,
        }
    }
    pub const fn from_bitsize_int(size: u32) -> Type {
        match size {
            ..=8 => Self::Byte,
            9..=16 => Self::Word,
            17..=32 => Self::DWord,
            33..=64 => Self::QWord,
            _ => unimplemented!(),
        }
    }
    pub const fn from_bytesize_int(size: u32) -> Type {
        Self::from_bitsize_int(size * 8)
    }

    pub const fn bits(self) -> u32 {
        match self {
            Type::Byte => 8,
            Type::Word => 16,
            Type::DWord => 32,
            Type::QWord => 64,
            Type::SinglePrecision => 32,
            Type::DoublePrecision => 64,
        }
    }

    pub const fn constant_inst(self) -> Inst {
        match self {
            Type::Byte => Inst::ConstantByte,
            Type::Word => Inst::ConstantWord,
            Type::SinglePrecision | Type::DWord => Inst::ConstantDWord,
            Type::DoublePrecision | Type::QWord => Inst::ConstantQWord,
        }
    }

    pub const fn bytes(self) -> u32 {
        match self {
            Type::Byte => 1,
            Type::Word => 2,
            Type::DWord => 4,
            Type::QWord => 8,
            Type::SinglePrecision => 4,
            Type::DoublePrecision => 8,
        }
    }
}

#[derive(Debug, PartialEq, Eq, Clone, Copy, Hash)]
pub enum Inst {
    /// index
    Label,
    // Constant(Type),
    /// index
    ConstantBytes,
    /// imm8
    ConstantByte,
    /// imm16
    ConstantWord,
    /// imm32
    ConstantDWord,
    /// imm64
    ConstantQWord,
    /// src
    LoadRegister(Type), // hint for loading value into register
    /// ast-node
    ExternRef,
    /// size, align
    Alloca,
    /// src
    Load(Type),
    /// src, dst
    Store(Type),
    /// src, idx,
    GetElementPtr(Type),
    Parameter(Type),
    /// lhs, rhs
    Add(Type),
    /// lhs, rhs
    Sub(Type),
    /// lhs, rhs
    Mul(Type),
    /// lhs, rhs
    MulSigned(Type),
    /// lhs, rhs
    Div(Type),
    /// lhs, rhs
    DivSigned(Type),
    /// lhs, rhs
    Rem(Type),
    /// lhs, rhs
    RemSigned(Type),
    /// lhs, rhs
    BitAnd(Type),
    /// lhs, rhs
    BitOr(Type),
    /// lhs, rhs
    BitXOr(Type),
    /// lhs
    Negate(Type),
    /// lhs, rhs
    ShiftLeft(Type),
    /// lhs, rhs
    ShiftRightSigned(Type),
    /// lhs, rhs
    ShiftRightUnsigned(Type),
    /// lhs
    ReturnValue,
    Return,
    //FunctionStart,
}

impl Inst {
    fn value_type(&self) -> Option<Type> {
        match self {
            Inst::Label
            | Inst::ConstantBytes
            | Inst::ExternRef
            | Inst::Alloca
            | Inst::ReturnValue
            | Inst::Store(_)
            | Inst::ConstantByte
            | Inst::ConstantWord
            | Inst::ConstantDWord
            | Inst::ConstantQWord
            | Inst::Return => None,
            Inst::GetElementPtr(ty)
            | Inst::Load(ty)
            | Inst::LoadRegister(ty)
            | Inst::Parameter(ty)
            | Inst::Add(ty)
            | Inst::Sub(ty)
            | Inst::Mul(ty)
            | Inst::MulSigned(ty)
            | Inst::Div(ty)
            | Inst::DivSigned(ty)
            | Inst::Rem(ty)
            | Inst::RemSigned(ty)
            | Inst::BitAnd(ty)
            | Inst::BitOr(ty)
            | Inst::BitXOr(ty)
            | Inst::Negate(ty)
            | Inst::ShiftLeft(ty)
            | Inst::ShiftRightSigned(ty)
            | Inst::ShiftRightUnsigned(ty) => Some(*ty),
        }
    }
    fn has_value(&self) -> bool {
        // basically, when an arithmetic instruction has two immediates, then just replace it with a mov into the dst reg
        // TODO: need to account for spilled values eventually; probably move this to `Mir`.
        match self {
            Inst::Label
            | Inst::ConstantBytes
            | Inst::ConstantByte
            | Inst::ConstantWord
            | Inst::ConstantDWord
            | Inst::ConstantQWord
            | Inst::ExternRef
            | Inst::Alloca
            | Inst::Store(_)
            | Inst::ReturnValue
            | Inst::Return => false,
            Inst::GetElementPtr(_)
            | Inst::Load(_)
            | Inst::LoadRegister(_)
            | Inst::Parameter(_)
            | Inst::Add(_)
            | Inst::Sub(_)
            | Inst::Mul(_)
            | Inst::MulSigned(_)
            | Inst::Div(_)
            | Inst::DivSigned(_)
            | Inst::Rem(_)
            | Inst::RemSigned(_)
            | Inst::BitAnd(_)
            | Inst::BitOr(_)
            | Inst::BitXOr(_)
            | Inst::Negate(_)
            | Inst::ShiftLeft(_)
            | Inst::ShiftRightSigned(_)
            | Inst::ShiftRightUnsigned(_) => true,
        }
    }
}

#[derive(Clone, Copy)]
pub union Data {
    none: (),
    imm8: u8,
    imm16: u16,
    imm32: u32,
    imm64: u64,
    index: StringsIndex,
    node: u32,
    binary: (u32, u32),
}

impl From<u8> for Data {
    fn from(value: u8) -> Self {
        Data::imm8(value)
    }
}
impl From<u16> for Data {
    fn from(value: u16) -> Self {
        Data::imm16(value)
    }
}
impl From<u32> for Data {
    fn from(value: u32) -> Self {
        Data::imm32(value)
    }
}
impl From<u64> for Data {
    fn from(value: u64) -> Self {
        Data::imm64(value)
    }
}

impl Data {
    pub fn imm8(v: u8) -> Data {
        Self { imm8: v }
    }
    pub fn imm16(v: u16) -> Data {
        Self { imm16: v }
    }
    pub fn imm32(v: u32) -> Data {
        Self { imm32: v }
    }
    pub fn imm64(v: u64) -> Data {
        Self { imm64: v }
    }
    pub fn index(v: StringsIndex) -> Data {
        Self { index: v }
    }
    pub fn node(v: u32) -> Data {
        Self { node: v }
    }
    pub fn binary(lhs: u32, rhs: u32) -> Data {
        Self { binary: (lhs, rhs) }
    }
    pub fn none() -> Data {
        Self { none: () }
    }

    pub fn as_imm8(self) -> u8 {
        unsafe { self.imm8 }
    }
    pub fn as_imm16(self) -> u16 {
        unsafe { self.imm16 }
    }
    pub fn as_imm32(self) -> u32 {
        unsafe { self.imm32 }
    }
    pub fn as_imm64(self) -> u64 {
        unsafe { self.imm64 }
    }
    pub fn as_index(self) -> StringsIndex {
        unsafe { self.index }
    }
    pub fn as_node(self) -> u32 {
        unsafe { self.node }
    }
    pub fn as_binary(self) -> (u32, u32) {
        unsafe { self.binary }
    }
}

use bitflags::bitflags;
bitflags! {
    #[derive(Debug, Clone, Copy)]
    struct BinaryOperandFlags: u8 {
        const RhsReg = 0b00000001;
        const RhsMem = 0b00000010;
        const RhsImm = 0b00000100;
        const LhsReg = 0b10000000;
        const LhsMem = 0b01000000;
        const LhsImm = 0b00100000;

        const RegReg = 0b10000001;
        const RegMem = 0b10000010;
        const RegImm = 0b10000100;
        const MemReg = 0b01000001;
        const MemMem = 0b01000010;
        const MemImm = 0b01000100;

        const RhsAll = 0b00000111;
        const LhsAll = 0b11100000;

        const NULL = 0b0;
    }
}
bitflags! {
    #[derive(Debug, Clone, Copy)]
    struct OperandKinds: u8 {
        const RegReg = 0b00000001;
        const RegMem = 0b00000010;
        const RegImm = 0b00000100;
        const MemReg = 0b00001000;
        const MemMem = 0b00010000;
        const MemImm = 0b00100000;
    }
}

impl OperandKinds {
    /// works for: add,sub,or,and,sbb,adc
    fn add() -> Self {
        Self::RegImm | Self::MemImm | Self::RegMem | Self::MemReg | Self::RegReg
    }
    /// imul is special...
    fn imul() -> Self {
        Self::RegImm | Self::MemImm | Self::RegMem | Self::RegReg
    }
    /// works for: div,idiv,mul
    fn mul() -> Self {
        Self::RegMem | Self::RegReg
    }
    /// works for: mulss,mulsd,divss,divsd,addss,addsd,subss,subsd
    fn sse() -> Self {
        Self::RegMem | Self::RegReg
    }
    /// works for: shl,shr,sar,sal
    fn shift() -> Self {
        Self::RegReg | Self::RegImm | Self::MemImm | Self::MemReg
    }
    const fn to_rhs_binop(self) -> BinaryOperandFlags {
        let reg = if self.intersects(Self::RegReg.union(Self::MemReg)) {
            BinaryOperandFlags::RhsReg
        } else {
            BinaryOperandFlags::empty()
        };
        let mem = if self.intersects(Self::RegMem.union(Self::MemMem)) {
            BinaryOperandFlags::RhsMem
        } else {
            BinaryOperandFlags::empty()
        };
        let imm = if self.intersects(Self::RegImm.union(Self::MemImm)) {
            BinaryOperandFlags::RhsImm
        } else {
            BinaryOperandFlags::empty()
        };

        reg.union(mem).union(imm)
    }
    const fn to_lhs_binop(self) -> BinaryOperandFlags {
        let reg = if self.intersects(Self::RegReg.union(Self::RegImm).union(Self::RegMem)) {
            BinaryOperandFlags::LhsReg
        } else {
            BinaryOperandFlags::empty()
        };
        let mem = if self.intersects(Self::MemReg.union(Self::MemMem).union(Self::MemImm)) {
            BinaryOperandFlags::LhsMem
        } else {
            BinaryOperandFlags::empty()
        };
        reg.union(mem)
    }

    fn reduce_with_rhs(self, lhs: BinaryOperandFlags) -> OperandKinds {
        let mut out = self;
        if !lhs.contains(BinaryOperandFlags::RhsImm) {
            out = out.difference(Self::MemImm | Self::RegImm);
        }
        if !lhs.contains(BinaryOperandFlags::RhsMem) {
            out = out.difference(Self::MemMem | Self::RegMem);
        }
        if !lhs.contains(BinaryOperandFlags::RhsReg) {
            out = out.difference(Self::RegReg | Self::MemReg);
        }
        out
    }
    fn reduce_with_lhs(self, lhs: BinaryOperandFlags) -> OperandKinds {
        let mut out = self;
        if !lhs.contains(BinaryOperandFlags::LhsMem) {
            out = out.difference(Self::MemReg | Self::MemMem | Self::MemImm);
        }
        if !lhs.contains(BinaryOperandFlags::LhsReg) {
            out = out.difference(Self::RegReg | Self::RegMem | Self::RegImm);
        }
        out
    }
}

enum OperandKind {
    Mem,
    Imm,
    Reg,
}

impl OperandKind {
    const fn as_rhs(self) -> BinaryOperandFlags {
        match self {
            OperandKind::Mem => BinaryOperandFlags::RhsMem,
            OperandKind::Imm => BinaryOperandFlags::RhsImm,
            OperandKind::Reg => BinaryOperandFlags::RhsReg,
        }
    }
    const fn as_lhs(self) -> BinaryOperandFlags {
        match self {
            OperandKind::Mem => BinaryOperandFlags::LhsMem,
            OperandKind::Imm => BinaryOperandFlags::LhsImm,
            OperandKind::Reg => BinaryOperandFlags::LhsReg,
        }
    }
}

struct BinaryOperands<'a> {
    mir: &'a mut Mir,
    commutative: bool,
    kinds: OperandKinds,
}

struct BinaryOperandsRunner<'a> {
    inner: BinaryOperands<'a>,
    lhs: (u32, Type),
    rhs: (u32, Type),
}

impl<'a> BinaryOperandsRunner<'a> {
    fn new(inner: BinaryOperands<'a>, lhs: u32, lhs_type: Type, rhs: u32, rhs_type: Type) -> Self {
        Self {
            inner,
            lhs: (lhs, lhs_type),
            rhs: (rhs, rhs_type),
        }
    }
    fn mir_mut(&mut self) -> &mut Mir {
        self.inner.mir
    }
    fn mir(&self) -> &Mir {
        self.inner.mir
    }
    fn lhs(&self) -> u32 {
        self.lhs.0
    }
    fn rhs(&self) -> u32 {
        self.rhs.0
    }
    fn lhs_type(&self) -> Type {
        self.lhs.1
    }
    fn rhs_type(&self) -> Type {
        self.rhs.1
    }
    fn lhs_rhs(&self) -> (u32, u32) {
        (self.lhs(), self.rhs())
    }
    fn canonicalise_lhs_with_reduced_kinds(&mut self, kinds: OperandKinds) {
        let (lhs, ty) = self.lhs;

        let l_legal = kinds.to_lhs_binop();
        let l_kind = self.mir().as_operand_kind(self.lhs()).as_lhs();

        if l_legal.contains(l_kind) {
        } else if l_legal.contains(BinaryOperandFlags::LhsReg) {
            self.lhs.0 = self.mir_mut().to_reg(ty, lhs);
        } else if l_legal.contains(BinaryOperandFlags::LhsMem) {
            self.lhs.0 = self.mir_mut().gen_spill_value(lhs);
        } else {
            unreachable!()
        }
    }

    fn try_swap(&mut self) {
        let (lhs, rhs) = self.lhs_rhs();
        let l_legal = self.inner.kinds.to_lhs_binop();
        let l_kind = self.mir().as_operand_kind(lhs).as_lhs();
        let r_kind = self.mir().as_operand_kind(rhs).as_rhs();

        if self.inner.commutative && (!l_legal.contains(l_kind) && l_legal.contains(r_kind)) {
            core::mem::swap(&mut self.lhs, &mut self.rhs);
        }
    }

    fn order(&mut self) {
        self.try_swap();
        let rhs = self.rhs();
        let ty = self.rhs_type();
        let r_legal = self.inner.kinds.to_rhs_binop();
        let r_kind = self.mir().as_operand_kind(rhs).as_rhs();

        if r_legal.contains(r_kind) {
        } else if r_legal.contains(BinaryOperandFlags::RhsReg) {
            self.rhs.0 = self.mir_mut().to_reg(ty, rhs);
        } else if r_legal.contains(BinaryOperandFlags::RhsMem) {
            self.rhs.0 = self.mir_mut().gen_spill_value(rhs);
        } else {
            unreachable!()
        }

        let rhs = self.rhs();
        self.canonicalise_lhs_with_reduced_kinds(
            self.inner
                .kinds
                .reduce_with_rhs(self.mir().as_operand_kind(rhs).as_rhs()),
        );
    }
}

impl<'a> BinaryOperands<'a> {
    fn new(mir: &'a mut Mir, commutative: bool, kinds: OperandKinds) -> Self {
        Self {
            mir,
            commutative,
            kinds,
        }
    }

    fn new_add_or_and_xor_adc(mir: &'a mut Mir) -> Self {
        Self::new(mir, true, OperandKinds::add())
    }
    fn new_sub_sbb(mir: &'a mut Mir) -> Self {
        Self::new(mir, false, OperandKinds::add())
    }
    fn new_sse(mir: &'a mut Mir) -> Self {
        Self::new(mir, true, OperandKinds::sse())
    }
    fn new_mul(mir: &'a mut Mir) -> Self {
        Self::new(mir, true, OperandKinds::mul())
    }
    fn new_imul(mir: &'a mut Mir) -> Self {
        Self::new(mir, true, OperandKinds::imul())
    }
    fn new_div_idiv_rem_irem(mir: &'a mut Mir) -> Self {
        Self::new(mir, false, OperandKinds::mul())
    }
    fn new_shift(mir: &'a mut Mir) -> Self {
        Self::new(mir, false, OperandKinds::shift())
    }

    fn wrangle(self, lhs: u32, lhs_type: Type, rhs: u32, rhs_type: Type) -> (u32, u32) {
        let mut runner = BinaryOperandsRunner {
            inner: self,
            lhs: (lhs, lhs_type),
            rhs: (rhs, rhs_type),
        };

        runner.order();

        runner.lhs_rhs()
    }
}

pub struct Mir {
    pub nodes: Vec<Inst>,
    pub data: Vec<Data>,
}

impl Mir {
    pub fn new() -> Mir {
        Self {
            nodes: Vec::new(),
            data: Vec::new(),
        }
    }

    fn as_operand_kind(&self, node: u32) -> OperandKind {
        if self.is_imm(node) {
            OperandKind::Imm
        } else if self.is_register(node) {
            OperandKind::Reg
        } else {
            OperandKind::Mem
        }
    }

    pub fn to_reg(&mut self, ty: Type, node: u32) -> u32 {
        if !self.is_register(node) {
            self.gen_load_register(ty, node)
        } else {
            node
        }
    }

    pub fn type_of_node(&self, node: u32) -> Option<Type> {
        self.nodes[node as usize].value_type()
    }

    pub fn is_register(&self, node: u32) -> bool {
        match self.nodes[node as usize] {
            Inst::LoadRegister(_)
            | Inst::Load(_)
            | Inst::GetElementPtr(_)
            | Inst::Parameter(_)
            | Inst::Add(_)
            | Inst::Sub(_)
            | Inst::Mul(_)
            | Inst::MulSigned(_)
            | Inst::Div(_)
            | Inst::DivSigned(_)
            | Inst::Rem(_)
            | Inst::RemSigned(_)
            | Inst::BitAnd(_)
            | Inst::BitOr(_)
            | Inst::BitXOr(_)
            | Inst::Negate(_)
            | Inst::ShiftLeft(_)
            | Inst::ShiftRightSigned(_)
            | Inst::ShiftRightUnsigned(_) => true,
            _ => false,
        }
    }

    pub fn is_imm(&self, node: u32) -> bool {
        match self.nodes[node as usize] {
            Inst::ConstantByte | Inst::ConstantWord | Inst::ConstantDWord | Inst::ConstantQWord => {
                true
            }
            _ => false,
        }
    }

    pub fn push(&mut self, inst: Inst, data: Data) -> u32 {
        let node = self.nodes.len() as u32;
        self.nodes.push(inst);
        self.data.push(data);

        node
    }

    pub fn gen_u8(&mut self, value: u8) -> u32 {
        self.push(Inst::ConstantByte, Data::imm8(value))
    }
    pub fn gen_u16(&mut self, value: u16) -> u32 {
        self.push(Inst::ConstantWord, Data::imm16(value))
    }
    pub fn gen_u32(&mut self, value: u32) -> u32 {
        self.push(Inst::ConstantDWord, Data::imm32(value))
    }
    pub fn gen_u64(&mut self, value: u64) -> u32 {
        self.push(Inst::ConstantQWord, Data::imm64(value))
    }
    pub fn gen_load_register(&mut self, ty: Type, src: u32) -> u32 {
        self.push(Inst::LoadRegister(ty), Data::node(src))
    }
    pub fn gen_spill_value(&mut self, src: u32) -> u32 {
        let ty = self.type_of_node(src).unwrap();
        let size = ty.bytes();
        let alloc = self.gen_alloca(size, size);
        _ = self.gen_store(ty, src, alloc);
        alloc
    }

    pub fn gen_label(&mut self, name: StringsIndex) -> u32 {
        self.push(Inst::Label, Data::index(name))
    }
    pub fn gen_alloca(&mut self, size: u32, align: u32) -> u32 {
        self.push(Inst::Alloca, Data::binary(size, align))
    }
    pub fn gen_load(&mut self, ty: Type, src: u32) -> u32 {
        self.push(Inst::Load(ty), Data::node(src))
    }
    pub fn gen_get_element_ptr(&mut self, ty: Type, src: u32, index: u32) -> u32 {
        self.push(Inst::GetElementPtr(ty), Data::binary(src, index))
    }
    pub fn gen_store(&mut self, ty: Type, src: u32, dst: u32) -> u32 {
        self.push(Inst::Store(ty), Data::binary(src, dst))
    }
    pub fn gen_param(&mut self, ty: Type) -> u32 {
        self.push(Inst::Parameter(ty), Data::none())
    }

    pub fn gen_truncate(&mut self, src: u32, ty: Type, signed: bool, bits: u16) -> u32 {
        // example: u4
        // ty = byte
        // 1111_1111 << 4 = 1111_0000
        // mask = 0000_1111;
        // shift = 8 - 4;
        let shift = ty.bits() as u8 - bits as u8;
        let mask = !(!0u64 << bits);

        let mask = match ty {
            Type::Byte => self.gen_u8(mask as u8),
            Type::Word => self.gen_u16(mask as u16),
            Type::SinglePrecision | Type::DWord => self.gen_u32(mask as u32),
            Type::DoublePrecision | Type::QWord => self.gen_u64(mask as u64),
        };

        let masked = self.gen_bitand(ty, src, mask);

        if signed {
            let shift = self.gen_u8(shift);
            let tmp = self.gen_shl(ty, masked, shift);
            let tmp = self.gen_sar(ty, tmp, shift);
            tmp
        } else {
            masked
        }
    }
    fn imm_to_reg(&mut self, src: u32) -> u32 {
        if self.is_imm(src) {
            // SAFETY: imms have values and thus types
            self.gen_load_register(self.type_of_node(src).unwrap(), src)
        } else {
            src
        }
    }

    pub fn gen_add(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        let (lhs, rhs) = if ty.is_floating() {
            BinaryOperands::new_sse(self).wrangle(lhs, ty, rhs, ty)
        } else {
            BinaryOperands::new_add_or_and_xor_adc(self).wrangle(lhs, ty, rhs, ty)
        };
        self.push(Inst::Add(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_sub(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        let (lhs, rhs) = if ty.is_floating() {
            BinaryOperands::new_sse(self).wrangle(lhs, ty, rhs, ty)
        } else {
            BinaryOperands::new_sub_sbb(self).wrangle(lhs, ty, rhs, ty)
        };
        self.push(Inst::Sub(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_mul(&mut self, ty: Type, signed: bool, lhs: u32, rhs: u32) -> u32 {
        if ty.is_floating() {
            self.gen_mul_sse(ty, lhs, rhs)
        } else if signed {
            self.gen_mul_signed(ty, lhs, rhs)
        } else {
            self.gen_mul_unsigned(ty, lhs, rhs)
        }
    }
    pub fn gen_mul_sse(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        let (lhs, rhs) = BinaryOperands::new_sse(self).wrangle(lhs, ty, rhs, ty);
        self.push(Inst::Mul(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_mul_unsigned(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        let (lhs, rhs) = BinaryOperands::new_mul(self).wrangle(lhs, ty, rhs, ty);
        self.push(Inst::Mul(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_mul_signed(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        let (lhs, rhs) = BinaryOperands::new_imul(self).wrangle(lhs, ty, rhs, ty);
        self.push(Inst::MulSigned(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_div(&mut self, ty: Type, signed: bool, lhs: u32, rhs: u32) -> u32 {
        if ty.is_floating() {
            self.gen_div_sse(ty, lhs, rhs)
        } else if signed {
            self.gen_div_signed(ty, lhs, rhs)
        } else {
            self.gen_div_unsigned(ty, lhs, rhs)
        }
    }
    pub fn gen_div_sse(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        let (lhs, rhs) = BinaryOperands::new_sse(self).wrangle(lhs, ty, rhs, ty);
        self.push(Inst::Div(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_div_unsigned(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        let (lhs, rhs) = BinaryOperands::new_div_idiv_rem_irem(self).wrangle(lhs, ty, rhs, ty);
        self.push(Inst::Div(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_div_signed(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        let (lhs, rhs) = BinaryOperands::new_div_idiv_rem_irem(self).wrangle(lhs, ty, rhs, ty);
        self.push(Inst::DivSigned(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_rem(&mut self, ty: Type, signed: bool, lhs: u32, rhs: u32) -> u32 {
        if ty.is_floating() {
            self.gen_rem_fp(ty, lhs, rhs)
        } else if signed {
            self.gen_rem_signed(ty, lhs, rhs)
        } else {
            self.gen_rem_unsigned(ty, lhs, rhs)
        }
    }
    pub fn gen_rem_fp(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        _ = (ty, lhs, rhs);
        todo!()
    }
    pub fn gen_rem_unsigned(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        let (lhs, rhs) = BinaryOperands::new_div_idiv_rem_irem(self).wrangle(lhs, ty, rhs, ty);
        self.push(Inst::Rem(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_rem_signed(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        let (lhs, rhs) = BinaryOperands::new_div_idiv_rem_irem(self).wrangle(lhs, ty, rhs, ty);
        self.push(Inst::RemSigned(ty), Data::binary(lhs, rhs))
    }

    pub fn gen_bitand(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        let (lhs, rhs) = BinaryOperands::new_add_or_and_xor_adc(self).wrangle(lhs, ty, rhs, ty);
        self.push(Inst::BitAnd(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_bitor(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        let (lhs, rhs) = BinaryOperands::new_add_or_and_xor_adc(self).wrangle(lhs, ty, rhs, ty);
        self.push(Inst::BitOr(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_bitxor(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        let (lhs, rhs) = BinaryOperands::new_add_or_and_xor_adc(self).wrangle(lhs, ty, rhs, ty);
        self.push(Inst::BitXOr(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_negate(&mut self, ty: Type, src: u32) -> u32 {
        let src = self.imm_to_reg(src);
        self.push(Inst::Negate(ty), Data::node(src))
    }

    #[doc(alias = "gen_shift_left")]
    pub fn gen_shl(&mut self, ty: Type, src: u32, shift: u32) -> u32 {
        let (src, shift) = BinaryOperands::new_shift(self).wrangle(src, ty, shift, ty);
        self.push(Inst::ShiftLeft(ty), Data::binary(src, shift))
    }
    #[doc(alias = "gen_shift_right")]
    pub fn gen_shr(&mut self, ty: Type, src: u32, shift: u32) -> u32 {
        let (src, shift) = BinaryOperands::new_shift(self).wrangle(src, ty, shift, ty);
        self.push(Inst::ShiftRightUnsigned(ty), Data::binary(src, shift))
    }
    #[doc(alias = "gen_shift_right_signed")]
    pub fn gen_sar(&mut self, ty: Type, src: u32, shift: u32) -> u32 {
        let (src, shift) = BinaryOperands::new_shift(self).wrangle(src, ty, shift, ty);
        self.push(Inst::ShiftRightSigned(ty), Data::binary(src, shift))
    }
    pub fn gen_ret_val(&mut self, val: u32) -> u32 {
        self.push(Inst::ReturnValue, Data::node(val))
    }
    pub fn gen_ret(&mut self) -> u32 {
        self.push(Inst::Return, Data::none())
    }
}

impl Mir {
    fn render_node<W: core::fmt::Write>(
        &self,
        w: &mut W,
        strings: &StringTable,
        reg_alloc: &BTreeMap<u32, amd64::Register>,
        node: u32,
    ) -> core::fmt::Result {
        let idx = node as usize;
        let inst = self.nodes[idx];
        let data = self.data[idx];

        if let Some(reg) = reg_alloc.get(&node) {
            write!(w, "({reg}) ")?;
        }

        match inst {
            Inst::Label => writeln!(w, "%{node} = {}", strings.get_str(data.as_index())),
            Inst::ConstantBytes => writeln!(
                w,
                "%{node} = bytes({:x?})",
                strings.get_bytes(data.as_index())
            ),
            Inst::ConstantByte => writeln!(w, "%{node} = imm8({:x?})", data.as_imm8()),
            Inst::ConstantWord => writeln!(w, "%{node} = imm16({:x?})", data.as_imm16()),
            Inst::ConstantDWord => writeln!(w, "%{node} = imm32({:x?})", data.as_imm32()),
            Inst::ConstantQWord => writeln!(w, "%{node} = imm64({:x?})", data.as_imm64()),
            Inst::LoadRegister(ty) => {
                let src = data.as_node();
                writeln!(w, "%{node} = load register {ty} %{src}")
            }
            Inst::ExternRef => writeln!(w, "%{node} = extern %%{}", data.as_node()),
            Inst::Alloca => {
                let (size, align) = data.as_binary();
                writeln!(w, "%{node} = alloca {size}, {align}")
            }
            Inst::GetElementPtr(ty) => {
                let (ptr, idx) = data.as_binary();
                writeln!(w, "%{node} = get element ptr {ty}, ptr %{ptr}, idx {idx}")
            }
            Inst::Load(ty) => {
                writeln!(w, "%{node} = load {ty}, ptr %{}", data.as_node())
            }
            Inst::Store(ty) => {
                let (src, dst) = data.as_binary();
                writeln!(w, "%{node} = store {ty}, ptr %{dst}, {ty} %{src}")
            }
            Inst::Parameter(ty) => {
                writeln!(w, "%{node} = param {ty}")
            }
            Inst::Add(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = add {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::Sub(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = sub {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::Mul(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = mul {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::MulSigned(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = signed mul {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::Div(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = div {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::DivSigned(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = signed div {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::Rem(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = rem {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::RemSigned(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = signed rem {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::BitAnd(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = bitand {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::BitOr(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = bitor {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::BitXOr(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = bitxor {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::Negate(ty) => {
                let lhs = data.as_node();
                writeln!(w, "%{node} = negate {ty} %{lhs}")
            }
            Inst::ShiftLeft(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = shift left {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::ShiftRightUnsigned(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = shift right {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::ShiftRightSigned(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = signed shift right {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::ReturnValue => {
                let lhs = data.as_node();
                writeln!(w, "%{node} = return %{lhs}")
            }
            Inst::Return => {
                writeln!(w, "%{node} = return")
            }
        }
    }

    pub fn render<W: core::fmt::Write>(
        &self,
        w: &mut W,
        strings: &StringTable,
    ) -> core::fmt::Result {
        let reg_alloc = self.build_liveness();
        for node in 0..self.nodes.len() {
            self.render_node(w, strings, &reg_alloc, node as u32)?;
        }
        Ok(())
    }

    pub fn display<'a, 'b>(&'a self, strings: &'b StringTable) -> DisplayMir<'a, 'b> {
        DisplayMir { mir: self, strings }
    }
}

impl Mir {
    pub fn build_liveness(&self) -> BTreeMap<u32, amd64::Register> {
        struct Interval {
            start: u32,
            end: u32,
        }

        let mut references = BTreeMap::<u32, Vec<u32>>::new();
        let mut preferred_colors = BTreeMap::<u32, amd64::Register>::new();
        use amd64::Register::*;
        let mut in_colors = [r9, r8, rcx, rdx, rdi, rsi].to_vec();
        let mut in_colors_sse = [xmm7, xmm6, xmm5, xmm4, xmm3, xmm2, xmm1, xmm0].to_vec();
        let mut inouts = Vec::<u32>::new();

        for i in 0..self.nodes.len() {
            let inst = self.nodes[i];
            let data = self.data[i];

            let node = i as u32;
            references.insert(node, Vec::new());

            match inst {
                Inst::Parameter(ty) => {
                    if let Some(reg) = if ty.is_floating() {
                        in_colors_sse.pop()
                    } else {
                        in_colors.pop()
                    } {
                        preferred_colors.insert(node, reg);
                    };
                    inouts.push(node);
                }
                // return is thru rax.
                Inst::ReturnValue => {
                    let val = data.as_node();
                    inouts.push(val);
                    references.get_mut(&val).unwrap().push(node);
                    _ = preferred_colors.try_insert(val, amd64::Register::rax);
                }
                Inst::Negate(_) | Inst::Load(_) => {
                    references.get_mut(&data.as_node()).unwrap().push(node);
                }
                Inst::GetElementPtr(_) => {
                    let (src, _) = data.as_binary();
                    references.get_mut(&src).unwrap().push(node);
                }
                // mul wants lhs (or rhs) to be rax, imul doesn't care as much.
                // note that it also clobers rdx
                Inst::Mul(ty) | Inst::MulSigned(ty) => {
                    let (lhs, rhs) = data.as_binary();
                    references.get_mut(&lhs).unwrap().push(node);
                    references.get_mut(&rhs).unwrap().push(node);

                    if !ty.is_floating() {
                        _ = preferred_colors.try_insert(lhs, amd64::Register::rax);
                    }
                }
                // div wants lhs to be rax, idiv can do either.
                // note that it also clobers rdx
                Inst::Div(ty) | Inst::DivSigned(ty) => {
                    let (lhs, rhs) = data.as_binary();
                    references.get_mut(&lhs).unwrap().push(node);
                    references.get_mut(&rhs).unwrap().push(node);

                    if !ty.is_floating() {
                        _ = preferred_colors.try_insert(lhs, amd64::Register::rax);
                    }
                }
                // div wants lhs to be rax, idiv can do either.
                // note that it also clobers rax
                Inst::Rem(ty) | Inst::RemSigned(ty) => {
                    let (lhs, rhs) = data.as_binary();
                    references.get_mut(&lhs).unwrap().push(node);
                    references.get_mut(&rhs).unwrap().push(node);

                    if !ty.is_floating() {
                        _ = preferred_colors.try_insert(lhs, amd64::Register::rdx);
                    }
                }
                // shr,shl,sar want the shift to be in cl.
                Inst::ShiftLeft(_) | Inst::ShiftRightSigned(_) | Inst::ShiftRightUnsigned(_) => {
                    let (lhs, rhs) = data.as_binary();
                    references.get_mut(&lhs).unwrap().push(node);
                    references.get_mut(&rhs).unwrap().push(node);
                    _ = preferred_colors.try_insert(rhs, amd64::Register::rcx);
                }
                // add,adc,sub,sbb,or,and,xor and mov don't care much about their source registers
                Inst::Add(_)
                | Inst::Sub(_)
                | Inst::BitAnd(_)
                | Inst::BitOr(_)
                | Inst::BitXOr(_)
                | Inst::Store(_) => {
                    let (lhs, rhs) = data.as_binary();
                    references.get_mut(&lhs).unwrap().push(node);
                    references.get_mut(&rhs).unwrap().push(node);
                }
                _ => {}
            }
        }

        references.retain(|&node, refs| !refs.is_empty() && self.nodes[node as usize].has_value());
        preferred_colors.retain(|&node, _| self.nodes[node as usize].has_value());

        let intervals = references
            .iter()
            .map(|(node, refs)| {
                (
                    *node,
                    Interval {
                        start: *node,
                        end: refs.last().cloned().unwrap(),
                    },
                )
            })
            .collect::<BTreeMap<u32, Interval>>();

        let mut edges = HashSet::<(u32, u32)>::new();
        for (&node, interval) in &intervals {
            for (&other, _) in references.range(interval.start + 1..interval.end) {
                edges.insert((node, other));
            }
        }
        let inference_graph = petgraph::graph::UnGraph::<(), ()>::from_edges(edges.into_iter());

        #[derive(Debug, Default, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
        enum Color {
            #[default]
            Unassigned,
            Tentative(amd64::Register),
            Final(amd64::Register),
        }

        impl Color {
            fn color(self) -> Option<amd64::Register> {
                match self {
                    Color::Unassigned => None,
                    Color::Tentative(color) | Color::Final(color) => Some(color),
                }
            }
        }

        struct Colorizer<'a> {
            mir: &'a Mir,
            graph: petgraph::graph::UnGraph<(), ()>,
            colors: BTreeMap<u32, Color>,
            preferred: BTreeMap<u32, amd64::Register>,
        }

        impl<'a> Colorizer<'a> {
            fn node_colors(&self, node: petgraph::graph::NodeIndex) -> &[amd64::Register] {
                let colors = if self.mir.nodes[node.index()]
                    .value_type()
                    .map(|t| t.is_floating())
                    == Some(true)
                {
                    &amd64::Register::SSE[..]
                } else {
                    &amd64::Register::GPR[..]
                };
                colors
            }

            fn prepass<I: IntoIterator<Item = u32>>(&mut self, inouts: I) {
                for node in inouts.into_iter() {
                    self.precolor_node(node.into());
                }

                let keys = self.preferred.keys().cloned().collect::<Vec<_>>();
                for node in keys {
                    self.precolor_node(node.into());
                }
            }

            fn precolor_node(&mut self, node: petgraph::graph::NodeIndex) {
                // prepass: assign preferred colours for in/out values and specific
                // instructions like mul/div which require one operand to be in rax
                let node_u32 = node.index() as u32;
                // only apply color here if we have a preference
                if let Some(preferred_color) = self.preferred.remove(&node_u32) {
                    let mut clique_colors = self
                        .graph
                        .neighbors(node)
                        .filter_map(|n| self.colors.get(&(n.index() as u32)).cloned());

                    if clique_colors
                        .find(|color| color.color() == Some(preferred_color))
                        .is_none()
                    {
                        self.colors
                            .insert(node_u32, Color::Tentative(preferred_color));
                    }
                };
                // .chain(self.node_colors(node).into_iter().cloned());
            }

            fn color_node(&mut self, node: petgraph::graph::NodeIndex) {
                // final pass:
                // look at clique colors and prefer to steal colors from
                // tentatively colored nodes. this results in preferential
                // coloring depending on the order of the prepass.
                let node_u32 = node.index() as u32;
                let clique_colors = self
                    .graph
                    .neighbors(node)
                    .filter_map(|n| self.colors.get(&(n.index() as u32)).cloned())
                    .collect::<BTreeSet<_>>();

                let colors = self
                    .node_colors(node)
                    .into_iter()
                    .filter(|&&r| !clique_colors.contains(&Color::Final(r)))
                    .cloned()
                    .collect::<BTreeSet<_>>();

                // eprintln!("coloring %{node_u32}:");
                // eprintln!("\twants: {:?}", self.colors.get(&node_u32));
                // eprintln!("\tclique: {clique_colors:?}");
                // eprintln!("\tcandidates: {colors:?}");

                match self.colors.entry(node_u32) {
                    Entry::Vacant(v) => {
                        // here we want to first check clique_colors with tentative coloring.
                        let color = colors
                            .into_iter()
                            .find_or_first(|&c| !clique_colors.contains(&Color::Tentative(c)))
                            .expect("ran out of registers :(");
                        v.insert(Color::Final(color));
                    }
                    Entry::Occupied(mut e) => {
                        // we prefer to steal
                        variant!(e.get() => &Color::Tentative(reg));
                        let color = colors
                            .into_iter()
                            .find_or_first(|&c| c == reg)
                            .expect("ran out of registers :(");
                        e.insert(Color::Final(color));
                    }
                }
            }

            fn finalise(self) -> BTreeMap<u32, amd64::Register> {
                self.colors
                    .into_iter()
                    .filter_map(|(node, c)| match c {
                        Color::Final(reg) => Some((node, reg)),
                        _ => None,
                    })
                    .collect()
            }
        }

        let mut colorizer = Colorizer {
            mir: self,
            graph: inference_graph,
            preferred: preferred_colors,
            colors: BTreeMap::new(),
        };

        // prepass: assign preferred colours for in/out values and specific
        // instructions like mul/div which require one operand to be in rax
        colorizer.prepass(inouts);

        for &node in references.keys().rev() {
            if !self.nodes[node as usize].has_value() {
                continue;
            }
            colorizer.color_node(node.into());
        }

        let colors = colorizer.finalise();

        // eprintln!(
        //     "Inference Graph:\n{:?}",
        //     petgraph::dot::Dot::with_attr_getters(
        //         &inference_graph,
        //         &[
        //             petgraph::dot::Config::EdgeNoLabel,
        //             petgraph::dot::Config::NodeNoLabel,
        //         ],
        //         &Box::new(|_, _| { "".to_owned() }),
        //         &Box::new(|_, node: (petgraph::graph::NodeIndex, &())| {
        //             let (node, _) = node;
        //             if let Some(reg) = assigned_colors.get(&(node.index() as u32)) {
        //                 format!("label=\"{} ({})\"", node.index(), reg)
        //             } else {
        //                 format!("label=\"{}\"", node.index())
        //             }
        //         }),
        //     ),
        // );

        colors
    }
}

use crate::asm::amd64::{self, Mnemonic, Operand, Operands};
use crate::variant;
type Asm = Vec<(Mnemonic, Operands)>;

#[allow(dead_code, unused)]
impl Mir {
    fn assemble(&self, strings: &StringTable) {
        let mut mapping = HashMap::<usize, (StringsIndex, usize)>::new();
        let mut entry = Asm::new();
        let mut branches = HashMap::<StringsIndex, Asm>::new();
        let mut stack_offset = 0;
        let mut current_branch = StringsIndex::none();
        branches.insert(current_branch, Asm::new());

        for i in 0..self.nodes.len() {
            let inst = self.nodes[i];
            let data = self.data[i];

            let branch = branches.get_mut(&current_branch).unwrap();

            match inst {
                Inst::Label => todo!(),
                Inst::ConstantBytes => todo!(),
                Inst::ConstantByte => todo!(),
                Inst::ConstantWord => todo!(),
                Inst::ConstantDWord => todo!(),
                Inst::ConstantQWord => todo!(),
                Inst::LoadRegister(ty) => todo!(),
                Inst::ExternRef => todo!(),
                Inst::Alloca => {
                    let (size, align) = data.as_binary();
                    let size = size.next_multiple_of(align);
                    stack_offset += size;
                    mapping.insert(i, (current_branch, branch.len()));
                    branch.push((Mnemonic::alloca, Operands::One(Operand::imm32(size))))
                }
                Inst::Load(ty) => {
                    // stuff
                    // branch.push((Mnemonic::mov, Operands::One(Operand::imm32(size))))
                }
                Inst::Store(ty) => todo!(),
                Inst::GetElementPtr(ty) => todo!(),
                Inst::Parameter(ty) => todo!(),
                Inst::Add(_) => todo!(),
                Inst::Sub(_) => todo!(),
                Inst::Mul(_) => todo!(),
                Inst::MulSigned(_) => todo!(),
                Inst::Div(_) => todo!(),
                Inst::DivSigned(_) => todo!(),
                Inst::Rem(_) => todo!(),
                Inst::RemSigned(_) => todo!(),
                Inst::BitAnd(_) => todo!(),
                Inst::BitOr(_) => todo!(),
                Inst::BitXOr(_) => todo!(),
                Inst::Negate(_) => todo!(),
                Inst::ShiftLeft(_) => todo!(),
                Inst::ShiftRightSigned(_) => todo!(),
                Inst::ShiftRightUnsigned(_) => todo!(),
                Inst::ReturnValue => todo!(),
                Inst::Return => todo!(),
            }
        }
    }
}

pub struct DisplayMir<'a, 'b> {
    mir: &'a Mir,
    strings: &'b StringTable,
}

impl<'a, 'b> core::fmt::Display for DisplayMir<'a, 'b> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.mir.render(f, &self.strings)
    }
}
